"""
Module to extract YOB from email using rule-based
"""
from itertools import product

import re
import pandas as pd
import numpy as np

from preprocessing_pgp.email.extractors.const import (
    DOB_FORMAT_DICT,
    DOB_REGEX_DICT,
    DOB_NUM_DIGIT_DICT
)


class EmailYOBExtractor:
    """
    Class contains logic to extract YOB from email
    """
    def __init__(self):
        self.yob_formats =\
            self.__generate_yob_dict()

    def __generate_yob_dict(self):
        date_list = [
            ['full_day', 'half_day', 'none'],  # DAY
            ['full_month', 'half_month', 'none'],  # MONTH
            ['full_year', 'half_year']  # YEAR
        ]

        possible_date_formats = list(product(*date_list))
        # Not possible if having day but not month
        possible_date_formats =\
            [date for date in possible_date_formats
             if (date[1] != 'none') | (date[0] == 'none')]
        # Remove case 4 number 2580 -> NOT YOB:1980
        possible_date_formats.remove(('half_day', 'half_month', 'half_year'))

        num_digits = []
        regexps = []
        formats = []

        for date_fm in possible_date_formats:
            date_num_digit = self.__get_date_digit(date_fm)
            date_regex = self.__get_date_regex(date_fm)
            date_format = self.__get_date_format(date_fm)

            num_digits.append(date_num_digit)
            regexps.append(date_regex)
            formats.append(date_format)

        return zip(num_digits, regexps, formats)

    # ? HELPER FUNCTION
    def __get_date_digit(
        self,
        date_fm
    ) -> int:
        return sum(DOB_NUM_DIGIT_DICT.get(fm, 0) for fm in date_fm)

    def __get_date_regex(
        self,
        date_fm
    ) -> str:
        return ''.join([DOB_REGEX_DICT.get(fm, '') for fm in date_fm])

    def __get_date_format(
        self,
        date_fm
    ) -> str:
        return ''.join(DOB_FORMAT_DICT.get(fm, '') for fm in date_fm)

    def _get_number_digit(
        self,
        string: str
    ) -> int:
        """
        Retrieve number of digit in string
        """
        return sum(c.isdigit() for c in string)

    def _extract_yob_regex(
        self,
        email_name: str,
        regex: str
    ) -> float:
        found_dates = re.findall(regex, email_name)

        if len(found_dates) == 0:
            return np.nan

    def _get_yob_with_format(
        self,
        email_name: str,
        yob_fm
    ) -> float:
        """
        Extract yob in email name with specific format
        """
        num_digit, regexp, _ = yob_fm

        # * Check for number of digit
        num_digit_name = self._get_number_digit(email_name)
        if num_digit_name != num_digit:
            return np.nan

        patterns = re.findall(regexp, email_name)

        if len(patterns) == 0:
            return np.nan

        # Case only year
        if isinstance(patterns[-1], str):
            yob = float(patterns[-1])
        else:
            yob = float(patterns[-1][-1])

        # Edge case 2K birth
        if yob == 20.0:
            return 2000.0

        # 1980 -> 1980, 8 -> 2008, 98 -> 1998
        yob = 1900+yob if 9<yob<100 else 2000+yob if yob <= 9 else yob

        return yob

    def _get_yob(
        self,
        data: pd.DataFrame,
        email_name_col: str = 'email_name'
    ) -> pd.DataFrame:
        """
        Generate an yob_extracted columns
        """
        data['yob_extracted'] = np.nan
        for yob_fm in self.yob_formats:
            not_found_mask = data['yob_extracted'].isna()
            data.loc[
                not_found_mask,
                'yob_extracted'
            ] = data.loc[
                not_found_mask,
                email_name_col
            ].apply(
                lambda email_name, fm=yob_fm:
                self._get_yob_with_format(email_name, fm)
            )

        return data


    # ? MAIN FUNCTION
    def extract_yob(
        self,
        data: pd.DataFrame,
        email_name_col: str = 'email_name'
    ) -> pd.DataFrame:
        """
        Extract YOB from email name

                Parameters
        ----------
        data : pd.DataFrame
            Data containing the email
        email_name_col : str, optional
            Name of email column from data, by default 'email_name'

        Returns
        -------
        pd.DataFrame
            Data with additional columns:
            * `yob_extracted` : The yob extracted from email
        """
        # * Extracting yob
        extracted_data = self._get_yob(
            data,
            email_name_col
        )

        return extracted_data



"""
This module meant for preprocessing (cleansing and unifying) the address before putting into extraction
"""

import re
from typing import Dict
from copy import deepcopy
from string import punctuation

import pandas as pd
from unidecode import unidecode

from preprocessing_pgp.address.utils import (
    number_pad_replace
)
from preprocessing_pgp.name.preprocess import (
    remove_spare_spaces,
)
from preprocessing_pgp.address.const import (
    DICT_NORM_ABBREV_REGEX_KW,
    DICT_NORM_CITY_DASH_REGEX,
    ADDRESS_PUNCTUATIONS
)


class VietnameseAddressCleaner:
    """
    Class support cleansing function for Vietnamese Addresses
    """
    non_address_punctuation = ''.join([pun for pun in punctuation
                                       if pun not in ADDRESS_PUNCTUATIONS])

    # * PRIVATE
    def __replace_with_keywords(self, address: str, keywords: Dict) -> str:
        """
        Helper function to replace sub-address with given dictionary of keywords

        Parameters
        ----------
        address : str
            The input address to replace with keywords
        keywords : Dict
            The dictionary with `str` key - 'target' and (list `str`) as value - 'keywords'

        Returns
        -------
        str
            The replaced address
        """

        replaced_address = deepcopy(address)

        for replace_txt, target_subs in keywords.items():
            reg_target = re.compile('|'.join(map(re.escape, target_subs)))
            replaced_address = re.sub(reg_target,
                                      replace_txt,
                                      replaced_address)

            if replaced_address != address:
                return replaced_address

        return replaced_address

    def __clean_address_with_regex(self, address: str, regex: str) -> str:
        """
        Helper function to clean any address with given non-grouping regex & removing spaces
        """
        address_match = re.search(regex, address)

        if address_match is not None:
            sub_address = address_match.group(0).strip()
            cleaned_address = address.replace(
                sub_address,
                sub_address.replace(' ', '')
            )
        else:
            cleaned_address = address

        return cleaned_address

    def __clean_digit_district(self, address: str) -> str:
        """
        Helper function to clean district with digit

        * E.g: 'p 7' -> 'p7', and more
        """
        district_regex = r'[^A|a]p [0-9]+'

        cleaned_address = self.__clean_address_with_regex(
            address, district_regex)

        return cleaned_address

    def __clean_digit_ward(self, address: str) -> str:
        """
        Helper function to clean ward with digit

        * E.g: 'q 7' -> 'q7', and more
        """
        ward_regex = r'q [0-9]+'

        cleaned_address = self.__clean_address_with_regex(address, ward_regex)

        return cleaned_address

    # * PROTECTED
    def _unify_address(self, address: str) -> str:
        """
        Helper function to unify address to lower words and unidecode
        """
        unified_address = deepcopy(address)
        unified_address = unified_address.lower()
        unified_address = unidecode(unified_address)

        return unified_address

    def _remove_spare_spaces(self, address: str) -> str:
        """
        Helper function to remove spare spaces from string

        Parameters
        ----------
        address : str
            The input address to remove spare spaces

        Returns
        -------
        str
            Clean address without any spare spaces
        """

        cleaned_address = re.sub(' +', ' ', address)
        cleaned_address = cleaned_address.strip()

        return cleaned_address

    def _remove_padding_number(self, address: str) -> str:
        """
        Helper function to remove any number in string with padding zeros
        """

        digit_group_regex = r'(\d+)'

        cleaned_address = re.sub(digit_group_regex,
                                 number_pad_replace,
                                 address)

        return cleaned_address

    def _clean_digit_address(self, address: str) -> str:
        """
        Helper function to clean address containing digits

        Parameters
        ----------
        address : str
            The raw address which may contain digits

        Returns
        -------
        str
            Clean address with digits
        """
        cleaned_address = self.__clean_digit_district(address)
        cleaned_address = self.__clean_digit_ward(cleaned_address)

        return cleaned_address

    def _clean_dash_address(self, address: str) -> str:
        """
        Helper function to clean cities with dash '-'

        * E.g: 'ba ria vung tau' -> 'ba ria - vung tau', and more
        """

        return self.__replace_with_keywords(address, DICT_NORM_CITY_DASH_REGEX)

    def _clean_abbrev_address(self, address: str) -> str:
        """
        Helper function to clean & unify abbrev in address
        """
        return self.__replace_with_keywords(address, DICT_NORM_ABBREV_REGEX_KW)

    def _clean_full_address(self, address: str) -> str:
        """
        Method for cleansing full address removing special characters

        Parameters
        ----------
        address : str
            The remained address to be cleaned

        Returns
        -------
        str
            The cleaned address
        """
        clean_address =\
            address.translate(str.maketrans(
                '', '', self.non_address_punctuation))

        clean_address = remove_spare_spaces(clean_address)

        return clean_address

    # * PUBLIC
    def clean_address(self, address: str) -> str:
        """
        Method for cleansing and unifying address

        Parameters
        ----------
        address : str
            The raw address that need cleansing and unifying

        Returns
        -------
        str
            Unified and cleaned address
        """
        unified_address = self._unify_address(address)

        cleaned_address = self._clean_abbrev_address(unified_address)

        cleaned_address = self._remove_spare_spaces(cleaned_address)

        cleaned_address = self._remove_padding_number(cleaned_address)

        cleaned_address = self._clean_digit_address(cleaned_address)

        cleaned_address = self._clean_dash_address(cleaned_address)

        cleaned_address = self._clean_full_address(cleaned_address)

        return cleaned_address


def clean_vi_address(
    data: pd.DataFrame,
    address_col: str
) -> pd.DataFrame:
    """
    Function to clean and unify vietnamese address in data

    Parameters
    ----------
    data : pd.DataFrame
        Raw data containing the address
    address_col : str
        The raw address column that need cleansing and unifying

    Returns
    -------
    pd.DataFrame
        Final unified and cleansed data with new column named `cleaned_<address_col>`
    """
    cleaner = VietnameseAddressCleaner()

    cleaned_data = data

    cleaned_data[f'cleaned_{address_col}'] =\
        cleaned_data[address_col].apply(
            cleaner.clean_address
    )

    return cleaned_data



"""
Module to extract address from email using rule-based
"""

import re
import pandas as pd

from preprocessing_pgp.address.const import LOCATION_ENRICH_DICT


class EmailAddressExtractor:
    """
    Class contains logic to extract address from email
    """

    def __init__(self):
        self.__generate_norm_dict()
        self.norm_regex = '|'.join(self.location_norm_dict.keys())

    def __generate_norm_dict(self):
        if hasattr(self, 'location_norm_dict'):
            return

        location_norm = LOCATION_ENRICH_DICT.copy()

        city_mask = LOCATION_ENRICH_DICT['lv1_norm'].str.contains(r'\btinh')

        location_norm.loc[
            city_mask,
            'norm_city'
        ] = location_norm.loc[
            city_mask,
            'lv1_norm'
        ].str.extract(r'\s(.*)')[0]\
            .str.replace(r'\s+', '', regex=True)\
            .str.replace('-', '', regex=False)

        location_norm.loc[
            ~city_mask,
            'norm_city'
        ] = location_norm.loc[
            ~city_mask,
            'lv1_norm'
        ].str.extract(r'\s\S*\s(.*)')[0]\
            .str.replace(r'\s+', '', regex=True)\
            .str.replace('-', '', regex=False)

        self.location_norm_dict = location_norm.set_index('norm_city')[
            'lv1'].to_dict()
        # * Additional cities
        self.location_norm_dict['vungtau'] = self.location_norm_dict['bariavungtau']
        self.location_norm_dict['baria'] = self.location_norm_dict['bariavungtau']

    def _get_address(
        self,
        email_name: str
    ) -> str:
        """
        Get Level 1 address from email name
        """
        matches = re.findall(self.norm_regex, email_name)

        if len(matches) == 0:
            return None

        norm_city = matches[0]
        address = self.location_norm_dict.get(norm_city, None)

        return address

    def extract_address(
        self,
        data: pd.DataFrame,
        email_name_col: str = 'email_name'
    ) -> pd.DataFrame:
        """
        Extract address from email name if possible

        Parameters
        ----------
        data : pd.DataFrame
            The input data contains an email_name column
        email_name_col : str, optional
            The name of the column contains the email name, by default 'email_name'

        Returns
        -------
        pd.DataFrame
            Data with additional column for extracted phone:
            * `address_extracted` : Extracted address (city-level 1) from email name
        """

        # * Using regex to search for phone
        data['address_extracted'] =\
            data[email_name_col].apply(self._get_address)

        return data



"""Module provides utils functions for card validation"""

import re
from typing import List
from string import ascii_lowercase
from string import punctuation

import pandas as pd


def digit_to_year_string(digit: int) -> str:
    """
    Convert digit in range (0, 99) to 2 digits birth string

    Parameters
    ----------
    digit : int
        The digit representing dob, must be in range `0-99` else will be thrown error

    Returns
    -------
    str
        The string representing the birth year
    """

    if not 0 <= digit <= 99:
        raise AttributeError("DOB Digit must be in range 0 to 99!")

    birth_str = f'0{digit}' if digit < 10 else str(digit)

    return birth_str


def is_checker_valid(checkers: List) -> bool:
    """
    Check if any of the checker is valid

    Returns
    -------
    bool
        Whether any of the checker is valid
    """

    return any(checkers)


def remove_spaces(sentence: str) -> str:
    """
    Function to remove all spaces in sentence

    Parameters
    ----------
    sentence : str
        The input sentence

    Returns
    -------
    str
        The output sentence without any spacing
    """

    # Remove spaces in between
    sentence = re.sub(' +', '', sentence)
    sentence = sentence.strip()

    return sentence


def remove_special_characters(sentence: str) -> str:
    """
    Removing special characters in string

    Parameters
    ----------
    sentence : str
        The sentence to remove punctuation

    Returns
    -------
    str
        The clean sentence without any punctuation
    """

    translator = str.maketrans('', '', punctuation)

    return sentence.translate(translator)


def check_contain_all_digit(
    card_id: str
) -> bool:
    """
    Simple function to check if the card_id contains all decimal

    Parameters
    ----------
    card_id : str
        The input card id

    Returns
    -------
    bool
        Whether the card id contains all decimal number
    """
    return card_id.isdecimal()


def check_non_digit(
    card_df: pd.DataFrame,
    card_col: str
) -> pd.Series:
    """
    Check if card contains any non_digit or not

    Parameters
    ----------
    card_df : pd.DataFrame
        The input card id DF
    card_col : str
        The column containing card id

    Returns
    -------
    pd.Series
        Series to verify id card
    """

    regex_non_digit = "|".join(list(ascii_lowercase))

    non_clean_mask = card_df[card_col].str.contains(regex_non_digit)

    return non_clean_mask



"""
Module to extract type from name
"""

from time import time

import pandas as pd
from flashtext import KeywordProcessor

from preprocessing_pgp.name.preprocess import preprocess_df
from preprocessing_pgp.name.accent_typing_formatter import remove_accent_typing
from preprocessing_pgp.utils import (
    parallelize_dataframe,
)
from preprocessing_pgp.name.type.const import (
    NAME_TYPE_REGEX_DATA
)


class TypeExtractor:
    """
    Class contains function to support for type extraction from name
    """

    def __init__(self) -> None:
        self.available_levels = NAME_TYPE_REGEX_DATA.keys()
        self.__generate_type_kws()

    def __generate_type_kws(self):
        if hasattr(self, 'type_kws'):
            return
        self.type_kws = {}
        for level in self.available_levels:
            level_kws = KeywordProcessor(case_sensitive=True)

            level_kws.add_keywords_from_dict(
                NAME_TYPE_REGEX_DATA[level]
            )
            self.type_kws[level] = level_kws

    def extract_type(
        self,
        name: str,
        level: str = 'lv1'
    ) -> str:
        """
        Extract name type from input name

        Parameters
        ----------
        name : str
            The input name to extract type from

        level : str
            The level to search for keyword pattern from,
            by default 'lv1' -- Currently there are 2 level 'lv1' and 'lv2'

        Returns
        -------
        str
            The first type extract from the name
        """
        if not hasattr(self, 'type_kws'):
            self.__generate_type_kws()

        results = self.type_kws[level].extract_keywords(name)

        if not results:
            return 'customer'
        return results[0]


def format_names(
    data: pd.DataFrame,
    name_col: str = 'name'
) -> pd.DataFrame:
    """
    Format name with clean and encoded without accent

    Parameters
    ----------
    data : pd.DataFrame
        The input data contains the name records
    name_col : str, optional
        The column name in data holds the name records, by default 'name'

    Returns
    -------
    pd.DataFrame
        Final data contains additional columns:

        * `de_<name_col>` contains decoded & lowered name derived from `name`
    """
    clean_data = data
    clean_data[f'de_{name_col}'] = clean_data[name_col].apply(
        remove_accent_typing)
    clean_data[f'de_{name_col}'] = clean_data[f'de_{name_col}'].str.lower()

    return clean_data


def extract_ctype(
    data: pd.DataFrame,
    name_col: str = 'de_name',
    level: str = 'lv1'
) -> pd.DataFrame:
    """
    Perform name-type extraction from formatted name col

    Parameters
    ----------
    data : pd.DataFrame
        The original dataframe contains the formatted name col
    name_col : str, optional
        The formatted name col, by default 'de_name'
    level : str, optional
        The level to process type extraction, by default 'lv1'

    Returns
    -------
    pd.DataFrame
        The new data contains additional columns:

        * `customer_type` contains the type of customer extracted from name
    """
    if data.empty:
        return data

    extracted_data = data
    # # ? KWS
    # type_extractor = TypeExtractor()
    # extracted_data['customer_type'] = extracted_data[name_col]\
    #     .apply(
    #         lambda name: type_extractor.extract_type(name, level)
    # )

    # ? Regex
    extracted_data.loc[
        extracted_data[name_col].notna(),
        'customer_type'
    ] = 'customer'
    level_name_type_regex = NAME_TYPE_REGEX_DATA[level]
    for ctype in level_name_type_regex.keys():
        regex_ctype = '|'.join(level_name_type_regex[ctype])
        # Only work with the non-predict type
        ctype_mask = (extracted_data[name_col].str.contains(regex_ctype)) &\
            (extracted_data['customer_type'] == 'customer')
        extracted_data.loc[
            ctype_mask,
            'customer_type'
        ] = ctype

    return extracted_data


def process_extract_name_type(
    data: pd.DataFrame,
    name_col: str = 'name',
    level: str = 'lv1',
    n_cores: int = 1,
    logging_info: bool = True
) -> pd.DataFrame:
    """
    Extract types from name records inputted from data

    Parameters
    ----------
    data : pd.DataFrame
        The input data contains the name records
    name_col : str, optional
        The column name in data holds the name records, by default 'name'
    level : str, optional
        The level to process type extraction, by default 'lv1'
    n_cores : int
        The number of cores used to run parallel, by default 1 core will be used

    Returns
    -------
    pd.DataFrame
        Final data contains additional columns:

        * `customer_type` contains type of customer extracted from `name` column
    """
    orig_cols = data.columns

    # ? Preprocess data
    if logging_info:
        print(">>> Cleansing names: ", end='')
    start_time = time()
    data = parallelize_dataframe(
        data,
        preprocess_df,
        n_cores=n_cores,
        name_col=name_col,
        clean_name=False
    )
    clean_time = time() - start_time
    if logging_info:
        print(f"{int(clean_time)//60}m{int(clean_time)%60}s")

    # * Remove NaNs and select only name col
    na_data = data[data[name_col].isna()][[name_col]]
    cleaned_data = data[data[name_col].notna()][[name_col]]

    # ? Format name
    if logging_info:
        print(">>> Formatting names: ", end='')
    start_time = time()
    formatted_data = parallelize_dataframe(
        cleaned_data,
        format_names,
        n_cores=n_cores,
        name_col=name_col
    )
    format_time = time() - start_time
    if logging_info:
        print(f"{int(format_time)//60}m{int(format_time)%60}s")

    # ? Extract name type by kws
    if logging_info:
        print(">>> Extracting customer's type: ", end='')
    start_time = time()
    extracted_data = parallelize_dataframe(
        formatted_data,
        extract_ctype,
        n_cores=n_cores,
        level=level,
        name_col=f'de_{name_col}',
    )
    extract_time = time() - start_time
    if logging_info:
        print(f"{int(extract_time)//60}m{int(extract_time)%60}s")

    # ? Drop clean_name column
    extracted_data = extracted_data.drop(columns=[f'de_{name_col}'])

    # ? Combined with Na data
    new_cols = ['customer_type']
    na_data[new_cols] = None
    final_data = pd.concat([extracted_data, na_data])

    # ? Combined with the origin cols
    final_data = pd.concat([data[orig_cols], final_data[new_cols]], axis=1)

    return final_data



"""
Module to detect for special phones and meanings in phones
"""
from preprocessing_pgp.phone.const import (
    DICT_NEW_MOBI_PHONE_VENDOR,
    DICT_NEW_TELEPHONE_VENDOR,
    DICT_PHONE_TAIL_TYPE
)


def detect_mobi_phone_vendor(phone: str) -> str:
    """
    Detect valid mobi phone's vendor

    Parameters
    ----------
    phone : str
        Valid mobi phone

    Returns
    -------
    str
        Mobi phone vendor
    """

    if phone[:3] in DICT_NEW_MOBI_PHONE_VENDOR.keys():
        return DICT_NEW_MOBI_PHONE_VENDOR[phone[:3]]

    return None


def detect_tele_phone_vendor(phone: str) -> str:
    """
    Detect valid tele phone's vendor

    Parameters
    ----------
    phone : str
        Valid tele phone

    Returns
    -------
    str
        Tele phone vendor
    """
    if phone[:4] in DICT_NEW_TELEPHONE_VENDOR.keys():
        return DICT_NEW_TELEPHONE_VENDOR[phone[:4]]

    if phone[:3] in DICT_NEW_TELEPHONE_VENDOR.keys():
        return DICT_NEW_TELEPHONE_VENDOR[phone[:3]]

    return None


def get_phone_tail_type(
    phone: str,
    tail_size: int
) -> str:
    """
    Get the tail type of phone with the given size

    Parameters
    ----------
    phone : str
        The valid phone number
    tail_size : int
        The size of the tail

    Returns
    -------
    str
        The type of the phone tail
    """
    return DICT_PHONE_TAIL_TYPE.get(phone[-tail_size:], None)


def detect_meaningful_phone(phone: str) -> str:
    """
    Detect the type of meaningful phone

    Parameters
    ----------
    phone : str
        Valid phone number

    Returns
    -------
    str
        Type of meaningful phone
    """

    # ? 6-3 last numbers
    for tail_size in range(6, 2, -1):
        tail_num_type = get_phone_tail_type(phone, tail_size)
        if tail_num_type is not None:
            return tail_num_type

    return "Số Thường"



"""
File containing constants that are necessary for processing address extraction
"""

import os

import pandas as pd


# ? REGEX FOR ADDRESS
DICT_NORM_ABBREV_REGEX_KW = {
    'tp ': ['tp.', 'tp:'],
    'tt ': ['tt.', 'tt:'],
    'q ': ['q.', 'q:'],
    'h ': ['h.', 'h:'],
    'x ': ['x.', 'x:'],
    'p ': ['p.', 'p:']
}

DICT_NORM_CITY_DASH_REGEX = {
    'ba ria - vung tau': ['ba ria vung tau'],
    'br-vt': ['brvt'],
    'phan rang - thap cham': ['phan rang thap cham'],
    'pr-tc': ['prtc'],
    'thua thien hue': ['thua thien - hue']
}

ADDRESS_PUNCTUATIONS = ['-', '/']

# ? LEVEL METHODS
LV1_METHODS = ['lv1_norm', 'lv1_abbrev', 'lv1_prefix_im', 'lv1_nprefix_im']
LV2_METHODS = ['lv2_norm', 'lv2_abbrev', 'lv2_prefix_im', 'lv2_nprefix_im']
LV3_METHODS = ['lv3_norm', 'lv3_abbrev', 'lv3_prefix_im', 'lv3_nprefix_im']
METHOD_REFER_DICT = {
    1: LV1_METHODS,
    2: LV2_METHODS,
    3: LV3_METHODS
}

# ? LOCATION ENRICH DICTIONARY
__LOCATION_ENRICH_PATH = os.path.join(
    os.path.dirname(__file__),
    os.pardir,
    "data",
    "location_dict_enrich_address.parquet"
)
LOCATION_ENRICH_DICT = pd.read_parquet(__LOCATION_ENRICH_PATH)

# ? LOCATION CODE DICTIONARY
__LOCATION_CODE_PATH = os.path.join(
    os.path.dirname(__file__),
    os.pardir,
    "data",
    "location_dict_code.parquet"
)
LOCATION_CODE_DICT = pd.read_parquet(__LOCATION_CODE_PATH)

LEVEL_VI_COLUMN_DICT = {
    1: 'city_vi',
    2: 'district_vi',
    3: 'ward_vi'
}
LEVEL_CODE_COLUMN_DICT = {
    1: 'city_id',
    2: 'district_id',
    3: 'ward_id'
}

AVAIL_LEVELS = LEVEL_VI_COLUMN_DICT.keys()



from setuptools import setup

# ? setup() does all the work
setup(
    include_dirs=['preprocessing_pgp/data'],
    include_package_data=True,
    package_data={
        "preprocessing_pgp.data": ["*"],
    }
)



"""
File support extracting code for each location extracted in extractor
"""
from typing import List, Dict

import pandas as pd

from preprocessing_pgp.address.const import (
    LOCATION_CODE_DICT,
    LEVEL_VI_COLUMN_DICT,
    LEVEL_CODE_COLUMN_DICT,
    AVAIL_LEVELS
)
from preprocessing_pgp.address.utils import (
    create_dependent_query,
    is_empty_string
)


class LocationCode:
    def __init__(self) -> None:
        self.loc_code_dict = LOCATION_CODE_DICT.copy()
        self.avail_levels = LEVEL_VI_COLUMN_DICT.keys()

        self.__unify_dictionary()

    def __unify_dictionary(self):
        """
        Unifying location code dict to titled names
        """
        for level in self.avail_levels:
            level_col = LEVEL_VI_COLUMN_DICT[level]
            self.loc_code_dict[level_col] = self.loc_code_dict[level_col].str.title(
            )

    def __get_level_col(
        self,
        level: int
    ) -> str:
        """
        Helper to get back the level column
        """
        return LEVEL_VI_COLUMN_DICT[level]

    def __get_level_code_col(
        self,
        level: int
    ) -> str:
        """
        Helper to get back the level code column
        """
        return LEVEL_CODE_COLUMN_DICT[level]

    def get_level_code(
        self,
        components: Dict,
    ) -> Dict:
        """
        Function to get the `code` of all possible levels

        * The provided `components` must have
        same length as the number of available levels (currently 3)

        Parameters
        ----------
        components : Dict
            `Dictionary` containing:
            * `key`: level
            * `value`: best_level_name

        Returns
        -------
        Dict
            The codes dictionary for each level
        """

        # * Create query to trace for location id
        component_queries = []
        for level in self.avail_levels:
            level_col = self.__get_level_col(level)
            best_lvl_name = components[level]
            if best_lvl_name is not None:
                level_query = f'{level_col} == "{best_lvl_name.title()}"'
                component_queries.append(level_query)

        trace_loc_id_query = create_dependent_query(*component_queries)

        # * Trace back location id
        level_codes = dict(
            zip(self.avail_levels,
                [None]*len(self.avail_levels))
        )
        if is_empty_string(trace_loc_id_query):
            return level_codes

        matches = self.loc_code_dict.query(trace_loc_id_query)

        if matches.shape[0] == 0:
            return level_codes

        for level in self.avail_levels:
            level_col = self.__get_level_code_col(level)
            match_code = matches[level_col].values[0]
            level_codes[level] = match_code

        # * Order the trace from lower level to higher level
        for level in self.avail_levels:
            if components[level] is None:
                for inv_level in reversed(list(self.avail_levels)):
                    if inv_level == level:
                        break
                    level_codes[inv_level] = None

                level_codes[level] = None
                break

        return level_codes


def generate_loc_code(
    data: pd.DataFrame,
    best_lvl_cols: List[str]
) -> pd.DataFrame:
    """
    Function to retrieve location code for each level:
    * `level_1_code`: code for city
    * `level_2_code`: code for district
    * `level_3_code`: code for ward

    Parameters
    ----------
    data : pd.DataFrame
        Raw data containing the address
    best_lvl_cols : List[str]
        The best extracted level column to retrieve code

    Returns
    -------
    pd.DataFrame
        New dataframe with new columns representing level code
    """

    code_generator = LocationCode()

    generated_data = data

    row_codes = generated_data.apply(
        lambda row: code_generator.get_level_code(
            dict(zip(
                AVAIL_LEVELS,
                [row[col] for col in best_lvl_cols]
            ))
        ),
        axis='columns'
    )

    for level in AVAIL_LEVELS:
        generated_data[f'level_{level}_code'] = [
            code[level] for code in row_codes]

    return generated_data



import os

import pandas as pd
import numpy as np

from preprocessing_pgp.card.utils import digit_to_year_string


# * IMPORTANT CONSTANTS
CURRENT_YEAR = 22
OLDEST_YEAR_DRIVER_LICENSE = 45


# * PERSONAL IDENTIFICATION CARD
__OLD_PID_CODE_PATH = os.path.join(
    os.path.dirname(__file__), os.pardir, "data", "old_codes.parquet"
)
__NEW_PID_CODE_PATH = os.path.join(
    os.path.dirname(__file__), os.pardir, "data", "new_codes.parquet"
)
__REGION_PID_CODE_PATH = os.path.join(
    os.path.dirname(__file__), '../data/new_region_code.parquet'
)

POSSIBLE_GENDER_NUM = ['0', '1', '2', '3']
GENDER_NUM_TO_CENTURY = {
    '20': ['0', '1'],
    '21': ['2', '3']
}
OLD_PID_REGION_CODE_NUMS = pd.read_parquet(__OLD_PID_CODE_PATH)['code'].values
NEW_PID_REGION_CODE_NUMS = pd.read_parquet(__NEW_PID_CODE_PATH)['code'].values
_REGION_PID_CODE = pd.read_parquet(__REGION_PID_CODE_PATH)
_DRIVER_LICENSE_CODE = _REGION_PID_CODE.copy()
_DRIVER_LICENSE_CODE['code'] = _DRIVER_LICENSE_CODE['code'].str[1:]
REGION_CODE_DICT = pd.concat([
    _REGION_PID_CODE,
    _DRIVER_LICENSE_CODE
], ignore_index=True)
OLD_PID_CODE_LENGTH = 9
NEW_PID_CODE_LENGTH = 12
LIMIT_DOB_PID = 14
VALID_PID_21_CENTURY_DOB = [
    digit_to_year_string(dob)
    for dob in range(0, CURRENT_YEAR-LIMIT_DOB_PID+1)
]

# * PASSPORT CARD
PASSPORT_LENGTH = 8
PASSPORT_PATTERN = r'^[a-z]\d{7}$'

# * DRIVER LICENSE CARD
'''
Documentation: https://tuhocvachiase.com/y-nghia-day-12-so-tren-bang-lai-xe-the-pet-xx-y-zz-1234567/
'''
DRIVER_LICENSE_ID_REGION_CODES = np.array(
    [code[1:] for code in NEW_PID_REGION_CODE_NUMS], dtype=object)

INVALID_DRIVER_LICENSE_PASSING_YEAR =\
    [digit_to_year_string(pass_year)
        for pass_year
        in range(CURRENT_YEAR+1, OLDEST_YEAR_DRIVER_LICENSE)
     ]

DRIVER_LICENSE_LENGTH = 12
INVALID_DRIVER_LICENSE_FIRST_YEAR_CHAR = ["3"]
VALID_DRIVER_LICENSE_LAST_YEAR_CHAR = ["1", "2"]



import os
import re
from abc import ABC, abstractmethod
from time import time

import pandas as pd
from tqdm import tqdm

from preprocessing_pgp.card.preprocess import (
    clean_card_data
)
from preprocessing_pgp.card.utils import is_checker_valid
from preprocessing_pgp.utils import (
    sep_display,
    # apply_multi_process,
    # apply_progress_bar,
    extract_null_values,
    parallelize_dataframe
)
from preprocessing_pgp.card.const import (
    # Personal ID
    OLD_PID_CODE_LENGTH,
    NEW_PID_CODE_LENGTH,
    POSSIBLE_GENDER_NUM,
    OLD_PID_REGION_CODE_NUMS,
    NEW_PID_REGION_CODE_NUMS,
    REGION_CODE_DICT,
    GENDER_NUM_TO_CENTURY,
    VALID_PID_21_CENTURY_DOB,
    # Passport
    PASSPORT_LENGTH,
    PASSPORT_PATTERN,
    # Driver License
    DRIVER_LICENSE_ID_REGION_CODES,
    INVALID_DRIVER_LICENSE_PASSING_YEAR,
    DRIVER_LICENSE_LENGTH,
    INVALID_DRIVER_LICENSE_FIRST_YEAR_CHAR,
    VALID_DRIVER_LICENSE_LAST_YEAR_CHAR
)

tqdm.pandas()
PROCESSES = os.cpu_count() // 2


class CardValidator(ABC):
    """
    Abstract Class contains validating functions for validating card_id
    """

    @staticmethod
    def is_valid_gender(gender_code: str) -> bool:
        """
        Check if the gender code is of correct format

        Parameters
        ----------
        gender_code : str
            Gender code extracted from card_id

        Returns
        -------
        bool
            Whether the gender code is at valid form
        """
        return gender_code in POSSIBLE_GENDER_NUM

    @staticmethod
    def is_all_digit_card(card_id: str) -> bool:
        """
        Check if the card_id contains all digit or not

        Parameters
        ----------
        card_id : str
            The card id

        Returns
        -------
        bool
            Whether the card id contains all digit or not
        """
        return card_id.isdecimal()

    @staticmethod
    @abstractmethod
    def is_valid_card(card_id: str) -> bool:
        """
        Check if the card is valid by syntax or not

        Parameters
        ----------
        card_id : str
            The input card id

        Returns
        -------
        bool
            Whether the card is valid by syntax
        """
        raise NotImplementedError("Subclasses should implement this!")


class PersonalIDValidator(CardValidator):
    """
    Class to check for personal id syntax
    """

    @staticmethod
    def is_old_card(card_id: str) -> bool:
        """
        Helper function to validate old card length
        """
        return len(card_id) == OLD_PID_CODE_LENGTH

    @staticmethod
    def is_new_card(card_id: str) -> bool:
        """
        Helper function to validate new card length
        """
        return len(card_id) == NEW_PID_CODE_LENGTH

    @staticmethod
    def is_valid_old_card(card_id: str) -> bool:
        """
        Helper function to check if the card_id is of old card type
        """
        if not PersonalIDValidator.is_old_card(card_id):
            return False
        return any((card in OLD_PID_REGION_CODE_NUMS
                    for card in [card_id[:2], card_id[:3]]))

    @staticmethod
    def is_valid_new_card(card_id: str) -> bool:
        """
        Helper function to check if the card_id is of new card type

        New card type contains:
        1.
        """
        if not PersonalIDValidator.is_new_card(card_id):
            return False

        if card_id[:3] in NEW_PID_REGION_CODE_NUMS:
            gender_code = card_id[3]
            return (
                CardValidator.is_valid_gender(gender_code)
                and PersonalIDValidator.is_valid_range_birth(card_id)
            )

        return False

    @staticmethod
    def is_semi_correct_length(card_id: str) -> bool:
        """
        Helper function to check if the card_id length is still acceptable
        """
        return len(card_id) in (OLD_PID_CODE_LENGTH-1, NEW_PID_CODE_LENGTH-1)

    @staticmethod
    def is_valid_semi_card(card_id: str) -> bool:
        """
        Helper function to check if the card_id is of semi-correct length and is correct

        Steps
        -----

        1. Check the length of the card id
        2. Modify card id by adding '0' at the beginning and check for whether it is valid or not
        """
        if not PersonalIDValidator.is_semi_correct_length(card_id):
            return False

        modified_card_id = '0' + card_id
        return PersonalIDValidator.is_valid_old_card(modified_card_id)\
            or PersonalIDValidator.is_valid_new_card(modified_card_id)

    @staticmethod
    def is_valid_range_birth(card_id: str) -> bool:
        """
        Helper to check whether the dob in new card_id is in valid range or not
        """
        gender_code = card_id[3]
        birth_year = card_id[4:6]

        # Century 21
        if gender_code in GENDER_NUM_TO_CENTURY['21']:
            return birth_year in VALID_PID_21_CENTURY_DOB

        return True

    @staticmethod
    def is_valid_card(card_id: str) -> bool:
        """
        Check if the card is valid by personal id syntax or not

        The valid card must be `all digit` and `valid syntax` of old or new card

        Parameters
        ----------
        card_id : str
            The input card id

        Returns
        -------
        bool
            Whether the card is valid by personal id syntax
        """
        if not CardValidator.is_all_digit_card(card_id):
            return False

        return PersonalIDValidator.is_valid_new_card(card_id)\
            or PersonalIDValidator.is_valid_old_card(card_id)\
            or PersonalIDValidator.is_valid_semi_card(card_id)


class PassportValidator(CardValidator):
    """
    Class to check for passport id syntax
    """

    @staticmethod
    def is_valid_length(card_id: str) -> bool:
        """
        Helper function to validate the correct length of the card
        """
        return len(card_id) == PASSPORT_LENGTH

    @staticmethod
    def is_valid_syntax(card_id: str) -> bool:
        """
        Helper function to validate passport syntax:

        Steps
        -----

        1. Contains a starting character
        2. Next is 7 random digits
        """
        return bool(re.match(PASSPORT_PATTERN, card_id.lower()))

    @staticmethod
    def is_valid_card(card_id: str) -> bool:
        """
        Check if the card is valid by passport syntax or not

        Parameters
        ----------
        card_id : str
            The input card id

        Returns
        -------
        bool
            Whether the card is valid by passport syntax
        """
        return PassportValidator.is_valid_length(card_id)\
            and PassportValidator.is_valid_syntax(card_id)


class DriverLicenseValidator(CardValidator):
    """
    Class to check for driver license syntax
    """

    @staticmethod
    def is_valid_length(card_id: str) -> bool:
        """
        Helper function to validate the correct length of the card
        """
        return len(card_id) == DRIVER_LICENSE_LENGTH

    @staticmethod
    def is_valid_region_code(card_id: str) -> bool:
        """
        Function to check for valid region code in driver license card id
        """
        region_code = card_id[:2]

        return region_code in DRIVER_LICENSE_ID_REGION_CODES

    @staticmethod
    def is_valid_passing_year(card_id: str) -> bool:
        """
        Helper function to validate driver license passing year in card:
        """
        passing_year = card_id[3:5]

        return passing_year not in INVALID_DRIVER_LICENSE_PASSING_YEAR

    @staticmethod
    def is_valid_gender_code(card_id: str) -> bool:
        """
        Helper function to validate driver license passing year in card:
        """
        gender_code = card_id[2]

        return CardValidator.is_valid_gender(gender_code)

    @staticmethod
    def is_real_driver_license(card_id: str) -> bool:
        """
        Helper function to validate the card id
        in the case of the first 3 digits is in personal id region code
        """
        if PersonalIDValidator.is_valid_new_card(card_id):
            first_year_char = card_id[3]
            second_year_char = card_id[4]

            return (first_year_char not in
                    INVALID_DRIVER_LICENSE_FIRST_YEAR_CHAR)\
                and (second_year_char in
                     VALID_DRIVER_LICENSE_LAST_YEAR_CHAR)

        return True

    @staticmethod
    def is_valid_card(card_id: str) -> bool:
        """
        Check if the card is valid by driver license syntax or not

        Parameters
        ----------
        card_id : str
            The input card id

        Returns
        -------
        bool
            Whether the card is valid by passport syntax
        """
        if not CardValidator.is_all_digit_card(card_id):
            return False

        return DriverLicenseValidator.is_valid_length(card_id)\
            and DriverLicenseValidator.is_valid_region_code(card_id)\
            and DriverLicenseValidator.is_valid_gender_code(card_id)\
            and DriverLicenseValidator.is_valid_passing_year(card_id)\
            and DriverLicenseValidator.is_real_driver_license(card_id)


def parse_gender_from_card(
    profile: pd.DataFrame,
    card_col: str = 'card_id',
) -> pd.DataFrame:
    """
    Parsing gender from new pid and driver license

    Parameters
    ----------
    profile : pd.DataFrame
        The input DF containing card id
    card_col : str, optional
        The column name that contain card data, by default 'card_id'

    Returns
    -------
    pd.DataFrame
        Data with extra info:
        * `gender`: Gender extracted from new pid and driver license
    """
    gender_dict = {
        '0': 'M',
        '1': 'F',
        '2': 'M',
        '3': 'F'
    }
    # * Gender from new personal's id
    profile['card_gender'] = None
    new_pid_mask = (profile['is_personal_id'])\
        & (profile[card_col].apply(PersonalIDValidator.is_new_card))
    profile.loc[
        new_pid_mask,
        'gender'
    ] = profile[card_col].str[3].map(gender_dict)

    # * Gender from driver license
    driver_license_mask = (profile['is_driver_license'])\
        & (profile['gender'].notna())

    profile.loc[
        driver_license_mask,
        'gender'
    ] = profile[card_col].str[2].map(gender_dict)

    return profile


def parse_birthday_from_card(
    profile: pd.DataFrame,
    card_col: str = 'card_id',
) -> pd.DataFrame:
    """
    Parsing birthday from new personal id

    Parameters
    ----------
    profile : pd.DataFrame
        The input DF containing card id
    card_col : str, optional
        The column name that contain card data, by default 'card_id'

    Returns
    -------
    pd.DataFrame
        Data with extra info:
        * `year_of_birth`: YOB extracted from new pid
    """
    # * Can only get birthday from new personal's id
    new_pid_mask = (profile['is_personal_id'])\
        & (profile[card_col].apply(PersonalIDValidator.is_new_card))

    gender_post_century = {
        '0': '19',
        '1': '19',
        '2': '20',
        '3': '20'
    }
    profile.loc[
        new_pid_mask,
        'post_cent'
    ] = profile[card_col].str[3].map(gender_post_century)

    profile.loc[
        new_pid_mask,
        'year_of_birth'
    ] = profile['post_cent'] + profile[card_col].str[4:6]

    profile = profile.drop(
        columns=['post_cent']
    )

    return profile


def parse_city_from_card(
    profile: pd.DataFrame,
    card_col: str = 'card_id',
) -> pd.DataFrame:
    """
    Parsing city from new personal id and driver license

    Parameters
    ----------
    profile : pd.DataFrame
        The input DF containing card id
    card_col : str, optional
        The column name that contain card data, by default 'card_id'

    Returns
    -------
    pd.DataFrame
        Data with extra info:
        * `city`: City extracted from new pid and driver license
    """
    # * New personal ID
    new_pid_mask = (profile['is_personal_id'])\
        & (profile[card_col].apply(PersonalIDValidator.is_new_card))
    profile.loc[
        new_pid_mask,
        'code'
    ] = profile[card_col].str[0:3]

    # * Driver License
    driver_license_mask = (profile['is_driver_license'])\
        & (~profile['is_personal_id'])
    profile.loc[
        driver_license_mask,
        'code'
    ] = profile[card_col].str[0:2]

    # * Merge to get city
    profile = pd.merge(
        profile.set_index('code'),
        REGION_CODE_DICT.set_index('code'),
        left_index=True, right_index=True,
        how='left',
        sort=False
    )\
        .reset_index()\
        .drop(columns=['code'])

    return profile


def extract_card_info(
    profile: pd.DataFrame,
    card_col: str = 'card_id'
) -> pd.DataFrame:
    """
    Parsing information from valid card id

    Parameters
    ----------
    profile : pd.DataFrame
        The input DF containing card id
    card_col : str, optional
        The column name that contain card data, by default 'card_id'

    Returns
    -------
    pd.DataFrame
        Data include extra info:
        * `gender`: Gender parsed
        * `year_of_birth`: YOB parsed
        * `city`: City parsed
    """
    if profile.empty:
        return profile

    # * Parsing gender from card id
    verified_data = parse_gender_from_card(
        profile,
        card_col=card_col
    )

    # * Parsing birthday from card id
    verified_data = parse_birthday_from_card(
        verified_data,
        card_col=card_col
    )

    # * Parsing city from card_id
    verified_data = parse_city_from_card(
        verified_data,
        card_col=card_col
    )

    return verified_data


def verify_card(
    data: pd.DataFrame,
    card_col: str = "card_id",
    print_info: bool = True
) -> pd.DataFrame:
    """
    Verify whether the card ids are valid or not

    Parameters
    ----------
    card_df : pd.DataFrame
        The input DF containing card id
    card_col : str, optional
        The column contain card id, by default "card_id"
    print_info : bool, optional
        Whether to print the information of the run, by default True

    Returns
    -------
    pd.DataFrame
        The final DF contains the columns that verify whether the card id is valid or not
    """
    if data.empty:
        return data

    # * Check for valid personal card id
    data['is_personal_id'] =\
        data[f"clean_{card_col}"]\
            .apply(PersonalIDValidator.is_valid_card)

    if print_info:
        print(f"# PERSONAL ID FOUND: {data['is_personal_id'].sum()}")
        sep_display()

    # * Check for valid passport id
    data['is_passport'] =\
        data[f"clean_{card_col}"]\
            .apply(PassportValidator.is_valid_card)

    if print_info:
        print(f"# PASSPORT FOUND: {data['is_passport'].sum()}")
        sep_display()

    # * Check for valid driver license id
    data['is_driver_license'] =\
        data[f"clean_{card_col}"].apply(DriverLicenseValidator.is_valid_card)

    if print_info:
        print(
            f"# DRIVER LICENSE FOUND: {data['is_driver_license'].sum()}")
        sep_display()

    # * Make a general is_valid column to verify whether the card is generally valid

    data['is_valid'] =\
        data.apply(
        lambda row: is_checker_valid(
            [
                row['is_personal_id'],
                row['is_passport'],
                row['is_driver_license']
            ]
        ),
        axis=1
    )

    return data


def process_verify_card(
    data: pd.DataFrame,
    card_col: str = "card_id",
    n_cores: int = 1
) -> pd.DataFrame:
    """
    Verify whether the card ids are valid or not with multi-core

    Parameters
    ----------
    card_df : pd.DataFrame
        The input DF containing card id
    card_col : str, optional
        The column contain card id, by default "card_id"
    n_cores : int, optional
        Number of cores to process, by default 1

    Returns
    -------
    pd.DataFrame
        The final DF contains the columns that verify whether the card id is valid or not
        * `is_valid`: Card is valid
        * `is_personal_id`: Card is personal id
        * `is_passport`: Card is passport
        * `is_driver_license`: Card is driver license
        * `gender`: Gender parsed
        * `year_of_birth`: YOB parsed
        * `city`: City parsed
    """
    orig_cols = data.columns
    card_data = data[[card_col]]

    # ? CLEAN CARD ID
    # * Removing na values
    clean_data, na_data = extract_null_values(card_data, card_col)

    # * Basic cleaning card_id
    print(">>> Cleansing card id: ", end='')
    start_time = time()
    clean_data = parallelize_dataframe(
        clean_data,
        clean_card_data,
        n_cores=n_cores,
        card_col=card_col
    )
    clean_time = time() - start_time
    print(f"Time takes {int(clean_time)//60}m{int(clean_time)%60}s")
    sep_display()

    # ? VALIDATE CARD ID
    print(">>> Verifying card id: ", end='')
    start_time = time()
    validated_data = parallelize_dataframe(
        clean_data,
        verify_card,
        n_cores=n_cores,
        card_col=card_col,
        print_info=False
    )
    verify_time = time() - start_time
    print(f"Time takes {int(verify_time)//60}m{int(verify_time)%60}s")
    sep_display()

    # ? PARSING INFO FROM CARD ID
    print(">>> Extracting card info: ", end='')
    start_time = time()
    parsed_data = parallelize_dataframe(
        validated_data,
        extract_card_info,
        n_cores=n_cores,
        card_col=f'clean_{card_col}'
    )
    extract_time = time() - start_time
    print(f"Time takes {int(extract_time)//60}m{int(extract_time)%60}s")
    sep_display()

    # ? CONCAT ALL SEP CARD IDS
    validator_cols = [
        'is_valid', 'is_personal_id',
        'is_passport', 'is_driver_license'
    ]
    new_cols = [
        f'clean_{card_col}',
        *validator_cols,
        'gender',
        'year_of_birth',
        'city'
    ]

    final_card_df = pd.concat([parsed_data, na_data], ignore_index=True)

    final_card_df[validator_cols] = final_card_df[validator_cols].fillna(False)

    final_card_df = final_card_df[[*orig_cols, *new_cols]]

    return final_card_df



import os
import json
import argparse
from time import time
from string import capwords
from typing import List
import multiprocessing as mp

import pandas as pd
from tqdm import tqdm
from unidecode import unidecode
from tensorflow import keras

from preprocessing_pgp.name.split_name import NameProcess
from preprocessing_pgp.name.model.transformers import TransformerModel
from preprocessing_pgp.name.rulebase_name import rule_base_name
from preprocessing_pgp.name.utils import remove_nicknames, is_name_accented
from preprocessing_pgp.utils import replace_trash_string

tqdm.pandas()


class NameProcessor:
    def __init__(self,
                 model: TransformerModel,
                 firstname_rb: pd.DataFrame,
                 middlename_rb: pd.DataFrame,
                 lastname_rb: pd.DataFrame,
                 ):
        self.model = model
        self.name_dicts = (firstname_rb, middlename_rb, lastname_rb)
        self.name_process = NameProcess()

    def choose_better_enrich(
        self,
        raw_name: str,
        enrich_name: str
    ) -> str:
        if raw_name is None or enrich_name is None:
            return None
        raw_components = raw_name.split(' ')
        enrich_components = enrich_name.split(' ')

        # * Enrich wrong
        if len(raw_components) != len(enrich_components):
            return None
        for i, component in enumerate(raw_components):
            if is_name_accented(enrich_components[i]) and not is_name_accented(component):
                return enrich_name

        return raw_name

    def predict_non_accent(self, name: str):
        if name is None:
            return None
        de_name = unidecode(name)

        # Keep case already have accent
        if name != de_name:
            return name

        # Only apply to case not having accent
        predicted_name = capwords(self.model.predict(de_name))

        return predicted_name

    def fill_accent(
        self,
        name_df: pd.DataFrame,
        name_col: str,
    ):
        orig_cols = name_df.columns

        # n_names = predicted_name.shape[0]
        # * Clean name before processing
        name_df[f'clean_{name_col}'] =\
            name_df[name_col].apply(
                lambda name: self.name_process.CleanName(name)[0]
        )

        # * Separate 1-word strange name
        one_word_mask = name_df[f'clean_{name_col}'].str.split(
            ' ').str.len() == 1
        strange_name_mask = (
            (one_word_mask) &
            (~self.name_process.check_name_valid(
                name_df[one_word_mask][f'clean_{name_col}']))
        )
        one_word_strange_names = name_df[strange_name_mask]
        normal_names = name_df[~strange_name_mask]

        # * Removing nicknames
        predicted_name = remove_nicknames(
            normal_names,
            name_col=f'clean_{name_col}'
        )

        predicted_name[['last_name', 'middle_name', 'first_name']] =\
            predicted_name[f'clean_{name_col}'].apply(
                self.name_process.SplitName).tolist()

        predicted_name[f'clean_{name_col}'] =\
            predicted_name[['last_name', 'middle_name', 'first_name']]\
            .fillna('').agg(' '.join, axis=1)\
            .str.strip()

        predicted_name = predicted_name.drop(
            columns=['last_name', 'middle_name', 'first_name']
        )

        # print("Filling diacritics to names...")
        # start_time = time()
        predicted_name['predict'] = predicted_name[f'clean_{name_col}'].apply(
            self.predict_non_accent)
        # mean_predict_time = (time() - start_time) / n_names

        # print(f"\nAVG prediction time : {mean_predict_time}s")

        # print('\n\n')

        # print("Applying rule-based postprocess...")
        # start_time = time()
        predicted_name['final'] = predicted_name.apply(
            lambda row: rule_base_name(
                row['predict'], row[f'clean_{name_col}'], self.name_dicts),
            axis=1
        )

        # predicted_name['final'] = predicted_name.apply(
        #     lambda row: self.choose_better_enrich(
        #         row[f'clean_{name_col}'], row['final']
        #     ),
        #     axis=1
        # )
        predicted_name[['last_name', 'middle_name', 'first_name']] =\
            predicted_name['final'].apply(
                self.name_process.SplitName).tolist()

        predicted_name['final'] = replace_trash_string(
            predicted_name,
            replace_col='final'
        )

        # * Full fill the data
        predicted_name = pd.concat([predicted_name, one_word_strange_names])

        # mean_rb_time = (time() - start_time) / n_names

        # print(f"\nAVG rb time : {mean_rb_time}s")

        # print('\n\n')
        out_cols = [
            'final', 'predict',
            'last_name', 'middle_name', 'first_name',
        ]

        return predicted_name[[*orig_cols, *out_cols]]

    def unify_name(self,
                   name_df: pd.DataFrame,
                   name_col: str,
                   key_col: str,
                   keep_cols: List[str]):
        name_df[key_col] = name_df[key_col].astype('str')
        best_name_df = self.name_process.CoreBestName(
            name_df,
            name_col=name_col,
            key_col=key_col
        )

        return best_name_df[[key_col, name_col, 'best_name', 'similarity_score', *keep_cols]]



import pandas as pd
import os

mobi_phone_path = os.path.join(
    os.path.dirname(__file__),
    os.pardir,
    'data',
    'mobi_head_code.parquet'
)
telephone_path = os.path.join(
    os.path.dirname(__file__),
    os.pardir,
    'data',
    'tele_head_code.parquet'
)

# ? PHONE EXTRACTION
sub_mobi_phone = pd.read_parquet(mobi_phone_path).reset_index()

sub_telephone = pd.read_parquet(telephone_path).reset_index()

SUB_PHONE_10NUM = sorted(sub_mobi_phone["NewSubPhone"].unique())
SUB_PHONE_11NUM = [
    x for x in sorted(sub_mobi_phone["OldSubPhone"].unique()) if len(x) == 4
]

STATIC_PHONE_HEAD = ["0218", "0219", "0210", "0211"]
SUB_TELEPHONE_10NUM = [
    x
    for x in sorted(sub_telephone["ma_vung_cu"].unique())
    if x not in STATIC_PHONE_HEAD
]
SUB_TELEPHONE_11NUM = sorted(sub_telephone["ma_vung_moi"].unique())

DICT_4_SUB_PHONE = (
    sub_mobi_phone[sub_mobi_phone["OldSubPhone"].str.len() == 4]
    .set_index("OldSubPhone")
    .to_dict()["NewSubPhone"]
)

DICT_4_SUB_TELEPHONE = (
    sub_telephone[~sub_telephone["ma_vung_cu"].isin(
        STATIC_PHONE_HEAD)]
    .set_index("ma_vung_cu")
    .to_dict()["ma_vung_moi"]
)

# ? PHONE VENDOR
DICT_NEW_MOBI_PHONE_VENDOR = (
    sub_mobi_phone.set_index('NewSubPhone')
    .to_dict()['PhoneVendor']
)

DICT_NEW_TELEPHONE_VENDOR = (
    sub_telephone.set_index('ma_vung_moi')
    .to_dict()['tinh']
)

# ? PHONE LENGTH
PHONE_LENGTH = {
    "old_mobi": 11,
    "new_mobi": 10,
    "old_landline": 10,
    "new_landline": 11
}

# ? MEANINGFUL PHONE
# TODO Change to syntax detection for meaningful phone
DICT_MEANINGFUL_PHONE = {
    'Lục Quý': ['000000', '111111', '222222', '333333', '444444',
                '555555', '666666', '777777', '888888', '999999'],
    'Ngũ Quý': ['00000', '11111', '22222', '33333', '44444',
                '55555', '66666', '77777', '88888', '99999'],
    'Tứ Quý': ['0000', '1111', '2222', '3333', '4444',
               '5555', '6666', '7777', '8888', '9999'],
    'Tam Hoa': ['000', '111', '222', '333', '444',
                '555', '666', '777', '888', '999'],
    'Số Tiến': ['0123', '1234', '2345', '3456', '4567', '5678', '6789'],
    'Số Lùi': ['3210', '4321', '5432', '6543', '7654', '8765', '9876'], # Low in money
    'Lộc Phát': ['6868', '6886', '8686', '8668',
                 '886', '866'],
    'Thần Tài': ['3939', '3993', '9339', '9393',
                 '7979', '7997', '9779', '9797',
                 '3979', '9397', '9379', '9739',
                 '7939', '9793'],
    'Ông Địa': ['3838', '3883', '8338', '8383',
                '7878', '7887', '8778', '8787',
                '3878', '8387', '8378', '8738',
                '7838', '8783']
}

DICT_PHONE_TAIL_TYPE = {}
for phone_type, tail_formats in DICT_MEANINGFUL_PHONE.items():
    for fm in tail_formats:
        DICT_PHONE_TAIL_TYPE[fm] = phone_type



"""
Module to extract phone from email using rule-based
"""

import re
import pandas as pd

from preprocessing_pgp.email.extractors.const import VIETNAMESE_PHONE_REGEX
from preprocessing_pgp.phone.extractor import process_convert_phone


class EmailPhoneExtractor:
    """
    Class contains logic to extract phone number from email
    """

    def _get_phone(
        self,
        email_name: str
    ) -> str:
        """
        Extract phone if exist from email name
        """

        phone_match = re.findall(VIETNAMESE_PHONE_REGEX, email_name)

        if len(phone_match) == 0:
            return None

        return phone_match[0][0]

    def extract_phone(
        self,
        data: pd.DataFrame,
        email_name_col: str = 'email_name'
    ) -> pd.DataFrame:
        """
        Extract phone number from email name if possible

        Parameters
        ----------
        data : pd.DataFrame
            The input data contains an email_name column
        email_name_col : str, optional
            The name of the column contains the email name, by default 'email_name'

        Returns
        -------
        pd.DataFrame
            Data with additional column for extracted phone:
            * `phone_extracted` : Extracted phone from email name
        """

        # * Using regex to search for phone
        data['phone_extracted'] =\
            data[email_name_col].apply(self._get_phone)

        # * Only take valid phone
        data['phone_extracted'] = process_convert_phone(
            data,
            phone_col='phone_extracted',
            logging_info=False
        )['phone_convert']

        return data
